//---------------------------------------------------------------------------


#pragma hdrstop

#include "WaveInput.h"
#include <stdio.h>

//---------------------------------------------------------------------------

#pragma package(smart_init)

WaveIn waveIn;

/** \brief Callback for WaveIn Device
*/
void CALLBACK WaveInProc(HWAVEIN waveOut, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	//The MM_WIM_DATA message is sent to a window when waveform-audio data
    //is present in the input buffer and the buffer is being returned to
    //the application. The message can be sent either when the buffer is full
    //or after the waveInReset function is called.
    //wParam = (WPARAM) hInputDev
    //lParam = (LONG) lpwvhdr
    if ( uMsg == MM_WIM_DATA )
    {
		waveIn.ProcessSamples();
    }
}

void __fastcall WaveIn::ProcessSamples()
{
//
//  Prepare data buffer
//
	if ( bRecordingStarted )
    {
        WAVEHDR *in = &WaveHeaders[iBuffIdx];
		//The waveInUnprepareHeader function cleans up the preparation performed
        //by the waveInPrepareHeader function. This function must be called
        //after the device driver fills a buffer and returns it to
        //the application. You must call this function before freeing the buffer.
        ::waveInUnprepareHeader( hWaveIn, in, sizeof(WAVEHDR) );

		// process samples
		if (!bStopRequest)
		{
			//if (bSamplesProcessed)
			//{
			//	bSamplesProcessed = false;
				if (callbackData)
					callbackData((short*)(in->lpData), WaveBufSize);
			//	bSamplesProcessed = true;
			//}
		}

		// next buffer to send
        iBuffIdx++;
        if ( iBuffIdx == iBuffersCount )
            iBuffIdx=0;
        SendHeader->dwBufferLength=WaveBufSize;
        SendHeader->dwFlags=0;
        SendHeader->dwLoops=0;
        ::waveInPrepareHeader( hWaveIn, SendHeader, sizeof(WAVEHDR) );
        ::waveInAddBuffer( hWaveIn, SendHeader, sizeof(WAVEHDR));
        if ( SendHeader == &WaveHeaders[iBuffersCount-1] )
            SendHeader = &WaveHeaders[0];
        else
            SendHeader++;
    }
}

WaveIn::WaveIn(void)
{
	bDeviceOpened = false;
	bRecordingStarted = false;
	bStopRequest = false;
	bSamplesProcessed = false;
	callbackData = NULL;
	iBuffersCount = 2;	// uses double-buffering
    iSamplesPerChannel = 1024; ///< total samples count (per channel)
    iBuffIdx = 0;
    waveFormat.wFormatTag = WAVE_FORMAT_PCM;
	waveFormat.nChannels = 1;
    waveFormat.nSamplesPerSec = 44100;
    waveFormat.wBitsPerSample = 16;
    waveFormat.nBlockAlign = waveFormat.wBitsPerSample/8*waveFormat.nChannels;
    waveFormat.nAvgBytesPerSec = waveFormat.nBlockAlign*waveFormat.nSamplesPerSec;
    waveFormat.cbSize = 0;
    //timethread = new DrawThread( true, iSamplesPerChannel );

	//  Audio buffers
	WaveHeaders = new WAVEHDR[iBuffersCount];
	WaveBufSize = iSamplesPerChannel * (waveFormat.wBitsPerSample/8) * waveFormat.nChannels;
    for ( int i = 0; i<iBuffersCount; i++ )
    {
        WaveHeaders[i].dwBufferLength = WaveBufSize;
        WaveHeaders[i].lpData = (char *)VirtualAlloc( 0, WaveBufSize, MEM_COMMIT, PAGE_READWRITE );
	}
}

WaveIn::~WaveIn()
{
	for ( int i=0; i<iBuffersCount; i++ )
        VirtualFree( WaveHeaders[i].lpData, 0, MEM_RELEASE );
    delete [] WaveHeaders;
}

/*
void WaveIn::EnumerateDevices(void)
{
	WAVEINCAPS     wic;

	TMenuItem *mi = new TMenuItem(miInput);
	mi->Caption = "Default (WAVE_MAPPER)";
	mi->Tag = 999;
	mi->OnClick = miSelectInputClick;
	mi->Checked = true;
	mi->RadioItem = true;
	miInput->Insert(0, mi);

	// Get the number of Digital Audio In devices in this computer
	int iNumDevs = waveInGetNumDevs();
	for (int i = 0; i < iNumDevs; i++)
	{
		// Get info about the next device
		if (!waveInGetDevCaps(i, &wic, sizeof(WAVEINCAPS)))
		{
			// Add device as menu item
			mi = new TMenuItem(miInput);
			mi->Caption = wic.szPname;
			mi->Tag = i;
			mi->OnClick = miSelectInputClick;
			mi->Checked = false;
			mi->RadioItem = true;
			miInput->Insert(i+1, mi);
		}
	}
}
*/

int WaveIn::Connect(int iDevId)
{
	if (bDeviceOpened)
		return 0;
	bRecordingStarted = false;
	if ( waveInOpen( &hWaveIn, WAVE_MAPPER, &waveFormat, (DWORD)WaveInProc, (DWORD)this, CALLBACK_FUNCTION ) == MMSYSERR_NOERROR)
	{
		for ( int i = 0; i<iBuffersCount/2; i++ )
		{
			WaveHeaders[i].dwBufferLength=WaveBufSize;
			WaveHeaders[i].dwFlags=0;
			WaveHeaders[i].dwLoops=0;
			if ( waveInPrepareHeader( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR))== MMSYSERR_NOERROR)
				if ( waveInAddBuffer( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR))!= MMSYSERR_NOERROR)
				{
					log("Error opening audio device."
						"Check if device is capable of recording with selected parameters or isn't busy.");
					return 1;
				}
		}
	}
	else
	{
		log("WaveIn: Multimedia Error");
		return 3;
	}
	bDeviceOpened = true;
	return 0;
}

int WaveIn::Disconnect(void)
{
	if (!bDeviceOpened)
		return 0;
	bRecordingStarted = false;

	// close wave device
	MMRESULT res=waveInReset( hWaveIn );
	for ( int i = 0; i<iBuffersCount; i++ )
	{
		res = ::waveInUnprepareHeader( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR) );
		while ( res == WAVERR_STILLPLAYING )
		{
			// The buffer pointed to by the pwh parameter is still in the queue.
			Sleep( 50 );
			res = ::waveInUnprepareHeader( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR) );
		}
	}
	::waveInClose( hWaveIn );

	bDeviceOpened = false;
	return 0;
}

int WaveIn::RecStart(void)
{
	bStopRequest = false;
	if ( waveInStart( hWaveIn ) == MMSYSERR_NOERROR )
	{
		SendHeader = &WaveHeaders[iBuffersCount/2];
		iBuffIdx = 0;
		bRecordingStarted = true;
	}
	else
	{
		log("WaveIn: Multimedia Error");
		return 1;
	}
	return 0;
}

int WaveIn::RecStop(void)
{
	bStopRequest = true;
	//Stops recording on a waveform audio input device and marks all pending data blocks as done
//	if ( waveInReset( hWaveIn ) != MMSYSERR_NOERROR )
//		log("WaveIn: waveInReset: Multimedia Error");
	if ( waveInStop( hWaveIn ) != MMSYSERR_NOERROR )
		log("WaveIn: waveInStop: Multimedia Error");
	return 0;
}

void WaveIn::log(char *lpData, ...)
{
	static char buf[256]; //determines max message length
	va_list ap;
	va_start(ap, lpData);
	vsnprintf(buf, sizeof(buf), lpData, ap);
	va_end(ap);
	buf[sizeof(buf)-1] = 0;

	if (callbackLog)
		callbackLog(buf);
}
